﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO.Ports;
using System.Threading;

namespace rae2_test
{
    public delegate void AddOutputWindowDelegate(string message);

    public class LogData
    {
        public DateTime time { get; set; }
        public UInt32 value { get; set; }
    }

    public class DoseRae2Communicator
    {
        private SerialPort port = null;
        private Object recvLock = new Object();

        List<byte> analyzeBuffer = new List<byte>();
        bool recieved = false;
        Packet lastpacket = null;

        AddOutputWindowDelegate AddOutputWindow = null;

        public DoseRae2Communicator(AddOutputWindowDelegate addOutputWindow)
        {
            AddOutputWindow = addOutputWindow;
        }

        ~DoseRae2Communicator()
        {
            this.PortClose();
        }

        private volatile bool stop = false;
        public void Stop()
        {
            stop = true;
        }

        public bool PortOpen(string comport)
        {
            if (port != null)
            {
                this.AddOutputWindow("ポートが既にオープンしています");
                return false;
            }

            if (comport == null)
            {
                this.AddOutputWindow("ポートが選択されていません");
                return false;
            }

            try
            {
                port = new SerialPort(comport, 38400);
                port.Open();
            }
            catch (Exception e)
            {
                port = null;
                this.AddOutputWindow("シリアルポートのオープンに失敗しました");
                return false;
            }

            port.DataReceived += new SerialDataReceivedEventHandler(DataRecievedHandler);
            this.AddOutputWindow(string.Format("シリアルポート({0:s})をオープンしました", comport));
            return true;
        }

        public bool PortClose()
        {
            if (port == null)
            {
                this.AddOutputWindow("シリアルポートはオープンしていません");
                return false;
            }

            port.Close();
            port = null;

            this.AddOutputWindow("シリアルポートをクローズしました");
            return true;
        }

        private void DataRecievedHandler(object sender, SerialDataReceivedEventArgs e)
        {
            if (stop)
            {
                return;
            }

            SerialPort sp = (SerialPort)sender;

            int readSize = sp.BytesToRead;
            byte[] indata = new byte[readSize];
            sp.Read(indata, 0, readSize);

            for (int i = 0; i < indata.Length; ++i)
            {
                byte c = indata[i];

                analyzeBuffer.Add(c);

                if (c == '}')
                {
                    Packet packet = Packet.Analyze(analyzeBuffer.ToArray());

                    lock (recvLock)
                    {
                        recieved = true;
                        lastpacket = packet;
                    }

                    if (packet == null)
                    {
                        this.AddOutputWindow("パケットの受け取りに失敗しました");
                    }
                    else if (packet.Command == 0x41)
                    {
                        if (packet.ContentsLength() == 32)
                        {
                            byte[] contents = packet.Contents;

                            int v1 = 0;
                            int p = 1;
                            v1 += contents[p++];
                            v1 = v1 * 0x100 + contents[p++];
                            v1 = v1 * 0x100 + contents[p++];
                            v1 = v1 * 0x100 + contents[p++];

                            p = 15;
                            int v2 = 0;
                            v2 += contents[p++];
                            v2 = v2 * 0x100 + contents[p++];
                            v2 = v2 * 0x100 + contents[p++];
                            v2 = v2 * 0x100 + contents[p++];

                            int v3 = 0;
                            v3 += contents[p++];
                            v3 = v3 * 0x100 + contents[p++];
                            v3 = v3 * 0x100 + contents[p++];
                            v3 = v3 * 0x100 + contents[p++];

                            int v4 = 0;
                            v4 += contents[p++];
                            v4 = v4 * 0x100 + contents[p++];

                            //this.AddOutputWindow(string.Format("{0:x} {1:d} {2:d} {3:x}", v1, v2, v3, v4));
                        }
                    }
                    analyzeBuffer.Clear();
                }
            }

            string asciiString = "";
            for (int i = 0; i < indata.Length; ++i)
            {
                char c = (char)indata[i];
                if (c == '{')
                {
                    asciiString += '\n';
                    asciiString += c;
                }
                else if (!char.IsControl(c))
                {
                    asciiString += c;
                }
                else
                {
                    asciiString += ".";
                }
            }

            //this.AddOutputWindow(asciiString);
        }

        private bool SendData(Packet packet)
        {
            if (port == null)
            {
                return false;
            }

            if (packet == null)
            {
                return false;
            }

            int contentsLength = packet.ContentsLength();
            if (contentsLength + 10 - 3 > 0xff)
            {
                return false;
            }

            byte[] senddata = new byte[contentsLength + 10];

            int p = 0;
            senddata[p++] = 0x7b;
            senddata[p++] = packet.Target;
            senddata[p++] = (byte)(contentsLength + 10 - 3);

            senddata[p++] = (byte)(packet.Value1 / 0x100);
            senddata[p++] = (byte)(packet.Value1 % 0x100);
            senddata[p++] = (byte)(packet.Value2 / 0x100);
            senddata[p++] = (byte)(packet.Value2 % 0x100);

            senddata[p++] = packet.Command;
            for (int i = 0; i < contentsLength; ++i)
            {
                senddata[p++] = packet.Contents[i];
            }

            // チェックサムの計算
            int sum = 0;
            for (int i = 0; i < p; ++i)
            {
                sum += senddata[i];
                sum = sum % 0x100;
            }
            if (sum != 0)
            {
                sum = 0x100 - sum;
            }
            senddata[p++] = (byte)sum;

            senddata[p++] = 0x7d;

            port.Write(senddata, 0, senddata.Length);
            return true;
        }

        private Packet SendAndRecvData(Packet sendPacket)
        {
            lock (this.recvLock)
            {
                this.recieved = false;
            }

            if (!this.SendData(sendPacket))
            {
                return null;
            }

            Packet recvPacket = null;
            bool r = true;
            DateTime stime = new DateTime(DateTime.Now.Ticks);
            stime = stime.AddSeconds(10);

            byte expectCommand;
            if (sendPacket.Target == 0x38)// && sendPacket.Command == 0x3a)
            {
                expectCommand = sendPacket.Command;
            }
            else
            {
                expectCommand = (byte)(sendPacket.Command + 1);
            }

            while (r && !stop)
            {
                lock (this.recvLock)
                {
                    if (this.recieved)
                    {
                        this.recieved = false;
                        recvPacket = lastpacket;
                        if (recvPacket != null && recvPacket.Command == expectCommand)
                        {
                            r = false;
                        }
                    }
                }

                if (DateTime.Compare(stime, DateTime.Now) < 0)
                {
                    this.AddOutputWindow("タイムアウトしました");
                    return null;
                }

                Thread.Sleep(1);
            }

            if (recvPacket == null)
            {
                return null;
            }

            return recvPacket;
        }

        /// <summary>
        /// デバイスと接続します
        /// </summary>
        /// <returns>接続できたらtrueを返します</returns>
        public bool CommandConnectDevice()
        {
            Packet sendPacket = new Packet(0x38, 0x0000, 0x0001, 0x3a, null);
            Packet recvPacket = this.SendAndRecvData(sendPacket);

            if (recvPacket == null)
            {
                return false;
            }

            if (recvPacket.ContentsLength() != 0xe4)
            {
                return false;
            }

            Packet sendPacket2 = new Packet(0x38, 0, 1, 0x3b, recvPacket.Contents);
            sendPacket2.Contents[0xaf] = 1;
            Packet recvPacket2 = this.SendAndRecvData(sendPacket2);

            if (recvPacket2 == null)
            {
                return false;
            }

            return true;
        }

        public bool CommandClearLog()
        {
            Packet sendPacket = new Packet(0x3c, 0x0001, 0x0001, 0x22, new byte[] { 0x88, 0xff });
            Packet recvPacket = this.SendAndRecvData(sendPacket);

            if (recvPacket == null)
            {
                return false;
            }

            this.AddOutputWindow("デバイスのログをクリアしました");
            return true;
        }

        public bool CommandClearD()
        {
            Packet sendPacket = new Packet(0x3c, 0x0001, 0x0001, 0x60, new byte[] { 0x5f });
            Packet recvPacket = this.SendAndRecvData(sendPacket);

            if (recvPacket == null)
            {
                return false;
            }

            this.AddOutputWindow("デバイスの累積値をクリアしました");
            return true;
        }

        public List<LogData> CommandGetAllData()
        {
            this.AddOutputWindow("全ログデータを取得します");

            int contentsLength;
            {
                Packet sendPacket = new Packet(
                    0x3c, 0x0001, 0x0001, 0x20, new byte[] { 0x83 });

                Packet recvPacket = this.SendAndRecvData(sendPacket);

                if (recvPacket == null || recvPacket.ContentsLength() != 5)
                {
                    return null;
                }
                int j = 1;
                contentsLength = recvPacket.Contents[j++];
                contentsLength = contentsLength * 0x100 + recvPacket.Contents[j++];
                contentsLength = contentsLength * 0x100 + recvPacket.Contents[j++];
                contentsLength = contentsLength * 0x100 + recvPacket.Contents[j++];
            }

            List<LogData> result = new List<LogData>();

            List<byte> contents = new List<byte>();

            for (int count = 0; ; ++count)
            {
                Packet sendPacket = new Packet(
                    0x3c, 0x0001, 0x0001, 0x20, new byte[] { 0x84, (byte)(count / 0x100), (byte)(count % 0x100) });

                Packet recvPacket = this.SendAndRecvData(sendPacket);

                if (recvPacket == null)
                {
                    return null;
                }

                if (recvPacket.ContentsLength() == 0)
                {
                    return null;
                }

                if (recvPacket.Contents[0] != 0x84)
                {
                    return null;
                }

                contents.AddRange(recvPacket.Contents.Skip(4));
                int length = recvPacket.Contents.Skip(4).Count();

                this.AddOutputWindow(string.Format("ログ読込中 {0:d}/{1:d}", count, length));

                if (recvPacket.Contents[1] == 1)
                {
                    break;
                }
            }

            if (contents.Count != contentsLength)
            {
                this.AddOutputWindow("ログの取得に失敗しました");
                return null;
            }

            int p = 0;
            while (p < contents.Count)
            {
                if (contents[p++] != 0xfd)
                {
                    return null;
                }
                if (contents[p++] != 0xfd)
                {
                    return null;
                }
                ++p; // 0x45
                p += 5; // serial
                p += 16; // unknown

                DateTime startTime;
                {
                    int year = 2000 + contents[p++];
                    int month = contents[p++];
                    int day = contents[p++];
                    int hour = contents[p++];
                    int min = contents[p++];
                    int sec = contents[p++];
                    startTime = new DateTime(year, month, day, hour, min, sec);
                }

                DateTime endTime;
                {
                    int year = 2000 + contents[p++];
                    int month = contents[p++];
                    int day = contents[p++];
                    int hour = contents[p++];
                    int min = contents[p++];
                    int sec = contents[p++];
                    endTime = new DateTime(year, month, day, hour, min, sec);
                }

                int sampleRate = contents[p] * 0x100 + contents[p + 1];
                p += 2;

                p += 21;

                int sampleCount = contents[p++];
                sampleCount = sampleCount * 0x100 + contents[p++];
                sampleCount = sampleCount * 0x100 + contents[p++];
                sampleCount = sampleCount * 0x100 + contents[p++];

                p += 8;
                p += 5; // serial
                p += 15;
                p += 4; // alarm
                p += 4; // alarm threthold
                p += 4; // over
                p += 1;

                DateTime time = new DateTime(startTime.Ticks);
                for (int i = 0; i < sampleCount; ++i)
                {
                    time = time.AddSeconds(sampleRate); // 最初のデータは開始時間ではない

                    UInt32 value = contents[p++];
                    value = value * 0x100 + contents[p++];
                    value = value * 0x100 + contents[p++];
                    value = value * 0x100 + contents[p++];

                    LogData logdata = new LogData();
                    logdata.time = time;
                    logdata.value = value;
                    result.Add(logdata);
                }
            }
            if (p != contents.Count)
            {
                this.AddOutputWindow("ログの読み込みで予期しない余りが出ました");
            }

            this.AddOutputWindow("ログを正常に取得しました");

            return result;
        }
    }

}
