﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace rae2_test
{
    class Packet
    {
        private byte[] contents;

        public byte Target { get; set; }
        public UInt16 Value1 { get; set; }
        public UInt16 Value2 { get; set; }
        public byte Command { get; set; }
        public byte[] Contents
        {
            get
            {
                return contents;
            }
            set
            {
                if (value == null)
                {
                    this.contents = value;
                }
                else if (value is byte[])
                {
                    this.contents = (byte[])value.Clone();
                }
                else
                {
                    throw new Exception();
                }
            }
        }

        public int ContentsLength()
        {
            if (this.contents == null)
            {
                return 0;
            }

            return this.contents.Length;
        }

        public Packet()
        {
            Target = 0;
            Value1 = 1;
            Value2 = 1;
            Command = 0;
            Contents = null;
        }

        public Packet(byte target, UInt16 value1, UInt16 value2, byte command, byte[] message)
        {
            this.Target = target;
            this.Value1 = value1;
            this.Value2 = value2;
            this.Command = command;
            this.Contents = message;
        }

        public static Packet Analyze(byte[] buffer)
        {
            if (buffer.Length < 10)
            {
                return null;
            }

            // チェックサムの確認
            int sum = 0;
            for (int i = 0; i < buffer.Length - 1; ++i)
            {
                sum += buffer[i];
                sum = sum % 0x100;
            }
            if (sum % 0x100 != 0)
            {
                return null;
            }

            Packet result = new Packet();

            int p = 0;

            if (buffer[p++] != 0x7b)
            {
                return null;
            }

            result.Target = buffer[p++]; // 38:Adapter 3C:Card

            int length = buffer[p++];
            if (length + 3 != buffer.Length)
            {
                return null;
            }

            // 1
            result.Value1 = (UInt16)(buffer[p] * 0x100 + buffer[p + 1]);
            p += 2;
            // 1
            result.Value2 = (UInt16)(buffer[p] * 0x100 + buffer[p + 1]);
            p += 2;

            result.Command = buffer[p++];

            if (length - 7 == 0)
            {
                result.Contents = null;
            }
            else
            {
                byte[] contents = new byte[length - 7];
                for (int i = 0; i < length - 7; ++i)
                {
                    contents[i] = buffer[p++];
                }
                result.Contents = contents;
            }

            int checksum = buffer[p++];
            if (buffer[p] != 0x7d)
            {
                return null;
            }

            return result;
        }
    }
}
